#include <iostream>
#include <stdio.h>
/** Quit the emulator. */
void quit(const char *why,int value1=0,int value2=0) 
	{ fprintf(stderr,why,value1,value2); exit(0); }

/** A tiny CPU Emulator */
class CPU {
public:
	enum {n_regs=16};
	int regs[n_regs]; ///< "registers"-- temporary working space of machine
	enum {n_mem=1024};
	int mem[n_mem]; ///< "memory"-- just program memory for now
	
	CPU() { /* set garbage values on machine boot */
		for (int r=0;r<n_regs;r++) regs[r]=0xDEADBEEF;
		for (int m=0;m<n_mem ;m++) mem [m]=0xD0A2BEEF;
	}
	int run(void);
	void dump(int id); /* Print out registers */
};

/*************** Main Emulator Run Routine *************/
int CPU::run(void) {
	while (1) {
		int inst=mem[regs[0xf]++]; /* instruction fetch */
		int opcode=(inst>>12)&0xf, i=(inst>>8)&0xf, 
		         j=(inst>> 4)&0xf, k=(inst>>0)&0xf;
		switch (opcode) { /* instruction decode */
		case 0x1: regs[i]=(inst&0xff); break; /* "load immediate" */
		case 0xA: regs[i]=regs[j]+regs[k]; break; /* Addition */
		case 0xB: if (regs[j]==regs[k]) regs[i]++; break; /* Cond */
		case 0xC: if (regs[j]>=regs[k]) regs[i]++; break; /* Cond */
		case 0xD: if (regs[j]>regs[k]) regs[i]++; break; /* Cond */
		case 0xE: /* Emulator (or OS) call: */
			switch (i) {
			case 0x0: printf("%d%c",regs[k]," \n\t,"[j]); break;
			case 0x1: printf("Enter a value:\n");  
				scanf("%i",&regs[k]); break; 
			case 0xD: dump(inst&0xff); break; 
			case 0xE: return regs[k]; /* Emulator Exit */
			default: quit("Unknown OS-trap instruction 0x%X at 0x%X!\n",inst,regs[0xf]-1);
			};
			break;
		default: quit("Unknown instruction 0x%X at 0x%X!\n",inst,regs[0xf]-1);
		};
	}
}

void CPU::dump(int id) {
	if (id==0xff) return; /* 0xEDFF is a disabled dump */
	printf("Dump %02x:\n",id);
	if (id<=0x80)
	  for (int m=0;m<n_mem;m++)
		if (mem[m]!=0)
			printf("Memory[%X]=0x%X \n",m,mem[m]);
			
	for (int i=0;i<16;i++)
		if (regs[i]!=0)
			printf("Register[%X]=0x%X  (or %d)\n",i,regs[i],regs[i]);
	printf("\n");
}

int main(int argc,char *argv[]) {
	CPU c;
	int initLoc=0x80; /* Location in memory where program is loaded */
	c.regs[0xf]=initLoc; /* Start at first instruction of program */
	
	/* Read program to run from the standard input */
	int instruction, loadCount=0;
	while (std::cin && std::cin>>std::hex>>instruction && instruction>=0) {
		c.mem[initLoc+loadCount++]=instruction;
		std::string comment; std::getline(std::cin,comment); /*skip comment*/
	}
	//c.dump(0x10); /* initial load */
	
	/* Run the program */
	int v=c.run();
	quit("Program complete.  Return code: %d (0x%X)\n",v,v);
	return 0;
}
